/*
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Based on the source code of PelDet written by Danzel A.
*/

#include <pspkernel.h>
#include <psppower.h>
#include <pspctrl.h>
#include <pspwlan.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <SDL.h>

#include "psp_global.h"
#include "psp_sdl.h"
#include "psp_danzeff.h"
#include "psp_menu.h"
#include "psp_fmgr.h"
#include "psp_irc.h"
#include "irc_global.h"

#include "psp_console.h"

# define CONSOLE_MAX_LINE      1024
# define CONSOLE_MAX_WIDTH       80

typedef unsigned char bool;

# define false      0
# define true       1

#define CONSOLE_DEFAULT_TOP_Y     0
#define CONSOLE_DEFAULT_BOTTOM_Y  (CONSOLE_SCREEN_HEIGHT-1)

# define BRIGHTNESS_NORMAL  0
# define BRIGHTNESS_BRIGHT  1
# define BRIGHTNESS_DIM     2

# define FONT_DEFAULT       0
# define FONT_LINEDRAW      1

  typedef struct console_elem_t {
     u8    m_char;
     u8    fg_color;
  } console_elem_t;

# define CONSOLE_SCREEN_TOP_Y  (CONSOLE_MAX_LINE - CONSOLE_SCREEN_HEIGHT)

  volatile SceUID consoleSema;

/* console screen */
static console_elem_t console_screen[CONSOLE_MAX_LINE][CONSOLE_MAX_WIDTH];
static int            console_colors[CONSOLE_MAX_COLOR];

static int  console_display_from_y = 0; 

static int loc_curr_x;
static int loc_curr_y;
static int loc_top_y    = CONSOLE_DEFAULT_TOP_Y;
static int loc_bottom_y = CONSOLE_DEFAULT_BOTTOM_Y;

static char loc_input_buffer[CONSOLE_SCREEN_WIDTH];
static int  loc_input_x = 0;

static void
psp_console_clear_line(int y)
{
  int x;
  for (x = 0; x < CONSOLE_SCREEN_WIDTH; x++) {
    console_elem_t *elem = &console_screen[CONSOLE_SCREEN_TOP_Y + y][x];
    elem->m_char = ' ';
    elem->fg_color = COLOR_WHITE;
  }
}


static void
psp_console_shift_up()
{
  int y;
	if (loc_top_y == 0) {	//if we're scrolling off the top of the screen then scroll the back buffer
    for (y = 0; y <= (CONSOLE_MAX_LINE-1-CONSOLE_SCREEN_HEIGHT); y++) {
      console_elem_t *scan_line1 = &console_screen[y    ][0];
      console_elem_t *scan_line2 = &console_screen[y + 1][0];
      memcpy(scan_line1, scan_line2, sizeof(console_elem_t) * CONSOLE_MAX_WIDTH);
    }
  }
	//now scroll the scroll area
  for (y = loc_top_y; y < loc_bottom_y; y++) {
    console_elem_t *scan_line1 = &console_screen[CONSOLE_SCREEN_TOP_Y + y    ][0];
    console_elem_t *scan_line2 = &console_screen[CONSOLE_SCREEN_TOP_Y + y + 1][0];
    memcpy(scan_line1, scan_line2, sizeof(console_elem_t) * CONSOLE_MAX_WIDTH);
  }
  psp_console_clear_line(y);
  loc_curr_y--;
}

static void
psp_console_shift_down()
{
  int y;
  for (y = loc_bottom_y; y > loc_top_y; y--) {
    console_elem_t *scan_line1 = &console_screen[CONSOLE_SCREEN_TOP_Y + y    ][0];
    console_elem_t *scan_line2 = &console_screen[CONSOLE_SCREEN_TOP_Y + y - 1][0];
    memcpy(scan_line1, scan_line2, sizeof(console_elem_t) * CONSOLE_MAX_WIDTH);
  }
  psp_console_clear_line(y);
  loc_curr_y++;
}

static void
psp_console_shift_up_if_needed()
{
  if (loc_curr_y > loc_bottom_y) {
    psp_console_shift_up();
  }
}

void
psp_console_clear_screen()
{
	sceKernelWaitSema(consoleSema, 1, 0);

  int x;
  int y;
  for (y = 0; y < CONSOLE_SCREEN_HEIGHT; y++) {
    for (x = 0; x < CONSOLE_SCREEN_WIDTH; x++) {
      console_elem_t *elem = &console_screen[CONSOLE_SCREEN_TOP_Y + y][x];
      elem->m_char = ' ';
      elem->fg_color = COLOR_WHITE;
    }
  }
  loc_curr_x = 0;
  loc_curr_y = 0;

	sceKernelSignalSema(consoleSema, 1);
}

void 
psp_console_write(char* buffer, int color)
{
	sceKernelWaitSema(consoleSema, 1, 0);

  int index;
  for (index = 0; buffer[index] != '\0'; index++) {
    char c = buffer[index];
    if (c == '\n') {
      loc_curr_y++;
      loc_curr_x = 0;
		  psp_console_shift_up_if_needed();
    } else 
    if (c == '\r') {
      loc_curr_x = 0;
    } else {
	    if (loc_curr_x >= CONSOLE_SCREEN_WIDTH) {
        loc_curr_y++;
        loc_curr_x = 0;
        psp_console_shift_up_if_needed();
	    }
      console_elem_t *elem = &console_screen[CONSOLE_SCREEN_TOP_Y + loc_curr_y][loc_curr_x];
      elem->m_char = c;
      elem->fg_color = color;
      loc_curr_x++;
    }
  }
	sceKernelSignalSema(consoleSema, 1);
}

void
psp_console_screen_scroll_up()
{
	sceKernelWaitSema(consoleSema, 1, 0);

  console_display_from_y -= CONSOLE_SCREEN_HEIGHT / 2;
  if (console_display_from_y < 0) console_display_from_y = 0;

	sceKernelSignalSema(consoleSema, 1);
}

void
psp_console_screen_scroll_down()
{
	sceKernelWaitSema(consoleSema, 1, 0);

  console_display_from_y += CONSOLE_SCREEN_HEIGHT / 2;
  if (console_display_from_y > CONSOLE_SCREEN_TOP_Y) console_display_from_y = CONSOLE_SCREEN_TOP_Y;

	sceKernelSignalSema(consoleSema, 1);
}

void
psp_console_display_screen()
{
	sceKernelWaitSema(consoleSema, 1, 0);

  psp_sdl_blit_console();

# if 0 //LUDO: FOR_DEBUG
  {
    char buffer[32];
    sprintf(buffer, "(%d %d)", loc_curr_x, loc_curr_y);
    int color = psp_sdl_rgb(0xff, 0xff, 0xff);
    psp_sdl_back_print(40, 6, buffer, color);
  }
# endif

  psp_sdl_select_font_6x10();

  int line = 0;
  int y_pix = 0;
  int x_pix = 0;

  int first_line = console_display_from_y;
  int last_line  = console_display_from_y + CONSOLE_SCREEN_HEIGHT;

  if (last_line >= CONSOLE_MAX_LINE) {
    first_line = CONSOLE_SCREEN_TOP_Y;
    last_line  = CONSOLE_SCREEN_TOP_Y + CONSOLE_SCREEN_HEIGHT - 1;
  }

  for (line = first_line; line <= last_line; line++) {
    console_elem_t *scan_line = &console_screen[line][0];
    x_pix = 0;
    int x;
    for (x = 0; x < CONSOLE_SCREEN_WIDTH; x++) {
      int fg_color = console_colors[scan_line[x].fg_color];
      char m_char = scan_line[x].m_char;
      if (m_char < ' ') m_char = ' ';
      psp_sdl_back_put_char(x_pix, y_pix, fg_color, m_char);
      x_pix += psp_font_width;
    }
    y_pix += psp_font_height;
  }

  y_pix = psp_font_height * CONSOLE_SCREEN_HEIGHT;
  x_pix = 0;
  int x = 0;
  for (x = 0; x < CONSOLE_SCREEN_WIDTH; x++) {
    char m_char;
    if (x >= loc_input_x) m_char = ' ';
    else                  m_char = loc_input_buffer[x];
    psp_sdl_back_put_char(x_pix, y_pix, console_colors[COLOR_WHITE], m_char);
    x_pix += psp_font_width;
  }
  x_pix = psp_font_width  * loc_input_x;
  psp_sdl_back_put_char(x_pix, y_pix, console_colors[COLOR_WHITE], '_');

  psp_sdl_select_font_8x8();

	sceKernelSignalSema(consoleSema, 1);
}

void
psp_console_init()
{
  int c;
  int c2;
  //c = 0xaF;
  //c = 0xFF;
  c = 0xf0;
  c2 = c >> 1;
  console_colors[COLOR_BLACK  ] = psp_sdl_rgb(0x00, 0x00, 0x00);
  console_colors[COLOR_RED    ] = psp_sdl_rgb(c   , 0x00, 0x00);
  console_colors[COLOR_GREEN  ] = psp_sdl_rgb(0x00, c   , 0x00);
  console_colors[COLOR_BLUE   ] = psp_sdl_rgb(0x00, 0x00, c   );
  console_colors[COLOR_YELLOW ] = psp_sdl_rgb(c   , c   , 0x00);
  console_colors[COLOR_MAGENTA] = psp_sdl_rgb(c   , c   , 0x00);
  console_colors[COLOR_GREY   ] = psp_sdl_rgb(c2  , c2  , c2  );
  console_colors[COLOR_CYAN   ] = psp_sdl_rgb(0x00, c   , c   );
  console_colors[COLOR_WHITE  ] = psp_sdl_rgb(c   , c   , c   );

  loc_curr_x = 0;
  loc_curr_y = 0;

  console_display_from_y = CONSOLE_SCREEN_TOP_Y;
  memset(console_screen, 0, sizeof(console_screen));

  //Semaphore
  if (! consoleSema) {
    consoleSema = sceKernelCreateSema("console", 0, 1, 1, NULL);
  }
}

void
psp_console_save_log(void)
{
	sceKernelWaitSema(consoleSema, 1, 0);

  FILE* File;
  char  TmpFileName[MAX_PATH];
  int   y = 0;
  int   x = 0;
  char  v = 0;

  sprintf(TmpFileName,"%s/log/log_%d.txt", PSPIRC.psp_homedir, PSPIRC.psp_log_id++);
  if (PSPIRC.psp_log_id >= 10) PSPIRC.psp_log_id = 0;
  File = fopen(TmpFileName, "w");
  if (File) {
    for (y = 0; y < CONSOLE_MAX_LINE; y++) {
      for (x = 0; x < CONSOLE_SCREEN_WIDTH; x++) {
        console_elem_t *elem = &console_screen[y][x];
        if (elem->m_char) break;
      }
      if (x != CONSOLE_SCREEN_WIDTH) break;
    }
    while (y < CONSOLE_MAX_LINE) {
      for (x = 0; x < CONSOLE_SCREEN_WIDTH; x++) {
        v = console_screen[y][x].m_char;
        if (v < ' ') v = ' ';
        fputc(v, File);
      }
      fputc('\n', File);
      y++;
    }
    fclose(File);
  }
	sceKernelSignalSema(consoleSema, 1);
}

void
psp_console_put_key(int key)
{
  char *scan_str   = NULL;
  int   length     = 0;
  int   max_length = 0;

  scan_str = loc_input_buffer;
  max_length = CONSOLE_SCREEN_WIDTH - 1;
  length = strlen(scan_str);

  if (key == DANZEFF_DEL) {
    if (length) {
      length--;
      scan_str[length] = '\0';
    }
  } else 
  if ((key == DANZEFF_ENTER) || (key == '\n')) {
    /* Send the line to the server */
    if (loc_input_x > 0) {
		  protocolSend(loc_input_buffer, FROM_USER);
      loc_input_buffer[0] = 0;
      loc_input_x = 0;
    }
    return;

  } else
  if (key >= ' ') {
  
    if ((length + 1) >= max_length) return;
    scan_str[length] = key;
    length++;
    scan_str[length] = 0;
  }
  loc_input_x = length;
}

void
psp_console_put_string(char *buffer)
{
  int index;
  for (index = 0; buffer[index]; index++) {
    psp_console_put_key(buffer[index]);
  }
}

